# SPDX-FileCopyrightText: 2016 Damien P. George
# SPDX-FileCopyrightText: 2017 Scott Shawcroft for Adafruit Industries
# SPDX-FileCopyrightText: 2019 Carter Nelson
# SPDX-FileCopyrightText: 2019 Roy Hooper
# SPDX-FileCopyrightText: 2022 Jeff Epler
#
# SPDX-License-Identifier: MIT

"""
`adafruit_neopxl8` - Neopixel strip driver using RP2040's PIO
=============================================================

* Author(s): Damien P. George, Scott Shawcroft, Carter Nelson, Roy Hooper, Jeff Epler
"""

import struct
import adafruit_pioasm
import bitops
import adafruit_pixelbuf
import rp2pio

_PROGRAM = """
.program piopixl8
top:
    mov pins, null      ; always-low part (last cycle is the 'pull ifempty' after wrap)
    pull block          ; wait for fresh data
    out y, 32           ; get count of NeoPixel bits

; NeoPixels are 800khz bit streams. We are choosing zeros as <312ns hi, 936 lo>
; and ones as <700 ns hi, 546 ns lo> and a clock of 16*800kHz, so the always-high
; time is 4 cycles, the variable time is 5 cycles, and the always-low time is 7 cycles
bitloop:
    pull ifempty [1]     ; don't start outputting HIGH unless data is available (always-low part)
    mov pins, ~ null [3] ; always-high part
    out pins, 8 [4]      ; variable part
    mov pins, null       ; always-low part (last cycle is the 'pull ifempty' after wrap)

    jmp y--, bitloop     ; always-low part

; A minimum delay is required so that the next pixel starts refreshing the front of the strands
    pull block
    out y, 32

wait_reset:
    jmp y--, wait_reset
    jmp top
"""

_ASSEMBLED = adafruit_pioasm.assemble(_PROGRAM)

# Pixel color order constants
RGB = "RGB"
"""Red Green Blue"""
GRB = "GRB"
"""Green Red Blue"""
RGBW = "RGBW"
"""Red Green Blue White"""
GRBW = "GRBW"
"""Green Red Blue White"""


class NeoPxl8(adafruit_pixelbuf.PixelBuf):
    """
    A sequence of neopixels.

    :param ~microcontroller.Pin data0: The first of 8 data registers, in GPIO order
    :param int n: The total number of neopixels.  Must be a multiple of the number of strands.
    :param int num_strands: The number of neopixels in each strand.
    :param int bpp: Bytes per pixel. 3 for RGB and 4 for RGBW pixels.
    :param float brightness: Brightness of the pixels between 0.0 and 1.0 where 1.0 is full
      brightness
    :param bool auto_write: True if the neopixels should immediately change when set. If False,
      `show` must be called explicitly.
    :param str pixel_order: Set the pixel color channel order. GRB or GRBW is
      set by default, depending on bpp.

    Example for Adafruit Feather RP2040 Scorpio, sets 240 LEDs in 8 strands to faint red:

    .. code-block:: python

        pixels = adafruit_neopxl8.NeoPxl8(board.NEOPIXEL0, 8*30, num_strands=8, auto_write=False)
        pixels.fill(0x010000)
        pixels.show()

    .. py:method:: NeoPxl8.show()

        Shows the new colors on the pixels themselves if they haven't already
        been autowritten. Note that with NeoPxl8 the show operation takes place
        in the background; when this routine returns, the new pixel data has just
        started to be written but your Python code can continue operating in the
        foreground, updating pixel values or performing other computations.

    .. py:method:: NeoPxl8.fill(color)

        Colors all pixels the given ***color***.

    .. py:attribute:: brightness

        Overall brightness of the pixel (0 to 1.0)

    """

    def __init__(
        self,
        data0,
        n,
        *,
        num_strands=8,
        bpp=3,
        brightness=1.0,
        auto_write=True,
        pixel_order=None,
    ):
        if n % num_strands:
            raise ValueError("Length must be a multiple of num_strands")
        if not pixel_order:
            pixel_order = GRB if bpp == 3 else GRBW
        else:
            if isinstance(pixel_order, tuple):
                order_list = [RGBW[order] for order in pixel_order]
                pixel_order = "".join(order_list)

        super().__init__(
            n, brightness=brightness, byteorder=pixel_order, auto_write=auto_write
        )

        data_len = bpp * n * 8 // num_strands
        padding_count = -data_len % 4

        self._data = bytearray(8 + data_len + padding_count)
        self._data32 = memoryview(self._data).cast("L")
        self._transposed = memoryview(self._data)[4 : 4 + data_len]
        self._data[:4] = struct.pack("<L", data_len - 1)
        self._data[-4:] = struct.pack("<L", 3840 * 2)

        self._num_strands = num_strands

        self._sm = rp2pio.StateMachine(
            _ASSEMBLED,
            frequency=800_000 * 16,
            first_out_pin=data0,
            out_pin_count=8,
            first_set_pin=data0,
            auto_pull=False,
            out_shift_right=True,
        )

    def deinit(self):
        """Blank out the neopixels and release the state machine."""
        self.fill(0)
        self.show()
        self._sm.deinit()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.deinit()

    def __repr__(self):
        return "[" + ", ".join([str(x) for x in self]) + "]"

    @property
    def n(self):
        """
        The total number of neopixels in all strands (read-only)
        """
        return len(self)

    @property
    def num_strands(self):
        """
        The total number of neopixels in all strands (read-only)
        """
        return self._num_strands

    def _transmit(self, buffer):
        while self._sm.pending:
            pass
        bitops.bit_transpose(buffer, self._transposed, self._num_strands)
        self._sm.background_write(self._data32)



# SPDX-FileCopyrightText: 2022 Jeff Epler
#
# SPDX-License-Identifier: Unlicense

import time
import board
from adafruit_neopxl8 import NeoPxl8

# Customize for your strands here
num_strands = 8
strand_length = 30
first_led_pin = board.NEOPIXEL0

num_pixels = num_strands * strand_length

# Make the object to control the pixels
pixels = NeoPxl8(
    first_led_pin,
    num_pixels,
    num_strands=num_strands,
    auto_write=False,
    brightness=1.00,
)

while True:
    for i in range(num_pixels):
        pixels[i] = 0x00_01_00
        pixels[i - 8] = 0x00_00_00
        pixels.show()
        time.sleep(1e-3)



# SPDX-FileCopyrightText: 2022 Jeff Epler
#
# SPDX-License-Identifier: Unlicense

# Set up a separate animation on each of 8 strips with NeoPxl8

import board
from adafruit_led_animation.animation.rainbow import Rainbow
from adafruit_led_animation.animation.comet import Comet
from adafruit_led_animation.animation.rainbowcomet import RainbowComet
from adafruit_led_animation.animation.rainbowchase import RainbowChase
from adafruit_led_animation.animation.chase import Chase
from adafruit_led_animation.group import AnimationGroup
from adafruit_led_animation.helper import PixelMap
from adafruit_led_animation import color
from adafruit_neopxl8 import NeoPxl8

# Customize for your strands here
num_strands = 8
strand_length = 30
first_led_pin = board.NEOPIXEL0

num_pixels = num_strands * strand_length

# Make the object to control the pixels
pixels = NeoPxl8(
    first_led_pin,
    num_pixels,
    num_strands=num_strands,
    auto_write=False,
    brightness=0.07,
)


def strand(i):
    return PixelMap(
        pixels,
        range(i * strand_length, (i + 1) * strand_length),
        individual_pixels=True,
    )


strands = [strand(i) for i in range(num_strands)]

animations = AnimationGroup(
    Comet(strands[0], 0.5, color.CYAN),
    Comet(strands[1], 0.4, color.AMBER),
    RainbowComet(strands[2], 0.3),
    RainbowComet(strands[3], 0.7),
    Chase(strands[4], 0.05, size=2, spacing=3, color=color.PURPLE),
    RainbowChase(strands[5], 0.05, size=2, spacing=3),
    Rainbow(strands[6], 0.6),
    Rainbow(strands[7], 0.23, step=21),
)

while True:
    animations.animate()
